from infi.hbaapi import Port
import json
example = json.loads(""" [
    {
        "discovered_ports": [
            {
                "discovered_ports": [], 
                "driver_name": null, 
                "driver_version": null, 
                "fabric_name": "00:00:00:00:00:00:00:00", 
                "firmware_version": null, 
                "hardware_version": null, 
                "manufacturer": null, 
                "model": null, 
                "model_description": null, 
                "node_wwn": "20:03:00:1b:32:6e:61:2e", 
                "option_rom_version": null, 
                "os_device_name": "", 
                "port_fcid": 5046528, 
                "port_max_frame_size": 0, 
                "port_speed": 0, 
                "port_state": "online", 
                "port_supported_speeds": [], 
                "port_symbolic_name": "", 
                "port_type": null, 
                "port_wwn": "21:03:00:1b:32:6e:61:2e", 
                "serial_number": null, 
                "statistics": {
                    "dumped_frames": -1, 
                    "error_frames": -1, 
                    "fcp_control_requests": -1, 
                    "fcp_input_megabytes": -1, 
                    "fcp_input_requests": -1, 
                    "fcp_output_megabytes": -1, 
                    "fcp_output_requests": -1, 
                    "invalid_crc_count": -1, 
                    "invalid_tx_word_count": -1, 
                    "link_failure_count": -1, 
                    "lip_count": -1, 
                    "loss_of_signal_count": -1, 
                    "loss_of_sync_count": -1, 
                    "nos_count": -1, 
                    "primitive_seq_protocol_err_count": -1, 
                    "rx_frames": -1, 
                    "rx_words": -1, 
                    "seconds_since_last_reset": -1, 
                    "tx_frames": -1, 
                    "tx_words": -1
                }
            }, 
            {
                "discovered_ports": [], 
                "driver_name": null, 
                "driver_version": null, 
                "fabric_name": "00:00:00:00:00:00:00:00", 
                "firmware_version": null, 
                "hardware_version": null, 
                "manufacturer": null, 
                "model": null, 
                "model_description": null, 
                "node_wwn": "20:00:00:1b:32:1e:0a:80", 
                "option_rom_version": null, 
                "os_device_name": "", 
                "port_fcid": 5046784, 
                "port_max_frame_size": 0, 
                "port_speed": 0, 
                "port_state": "online", 
                "port_supported_speeds": [], 
                "port_symbolic_name": "", 
                "port_type": null, 
                "port_wwn": "21:00:00:1b:32:1e:0a:80", 
                "serial_number": null, 
                "statistics": {
                    "dumped_frames": -1, 
                    "error_frames": -1, 
                    "fcp_control_requests": -1, 
                    "fcp_input_megabytes": -1, 
                    "fcp_input_requests": -1, 
                    "fcp_output_megabytes": -1, 
                    "fcp_output_requests": -1, 
                    "invalid_crc_count": -1, 
                    "invalid_tx_word_count": -1, 
                    "link_failure_count": -1, 
                    "lip_count": -1, 
                    "loss_of_signal_count": -1, 
                    "loss_of_sync_count": -1, 
                    "nos_count": -1, 
                    "primitive_seq_protocol_err_count": -1, 
                    "rx_frames": -1, 
                    "rx_words": -1, 
                    "seconds_since_last_reset": -1, 
                    "tx_frames": -1, 
                    "tx_words": -1
                }
            }
        ], 
        "driver_name": "ql2300.sys", 
        "driver_version": "9.1.8.6", 
        "fabric_name": "20:01:00:05:73:dc:fa:81", 
        "firmware_version": "4.06.01", 
        "hardware_version": "", 
        "manufacturer": "QLogic Corporation", 
        "model": "QLE2562", 
        "model_description": "QLogic QLE2562 Fibre Channel Adapter", 
        "node_wwn": "20:00:00:24:ff:2c:4e:53", 
        "option_rom_version": "2.16", 
        "os_device_name": "\\\\\\\\.\\\\Scsi6:", 
        "port_fcid": 5047040, 
        "port_max_frame_size": 2048, 
        "port_speed": 8, 
        "port_state": "online", 
        "port_supported_speeds": [
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4e:53", 
        "serial_number": "I07758", 
        "statistics": {
            "dumped_frames": -1, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": -1, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": -1, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 0, 
            "lip_count": -1, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": -1, 
            "primitive_seq_protocol_err_count": 0, 
            "rx_frames": -1, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": -1, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [
            {
                "discovered_ports": [], 
                "driver_name": null, 
                "driver_version": null, 
                "fabric_name": "00:00:00:00:00:00:00:00", 
                "firmware_version": null, 
                "hardware_version": null, 
                "manufacturer": null, 
                "model": null, 
                "model_description": null, 
                "node_wwn": "20:00:00:1b:32:17:63:91", 
                "option_rom_version": null, 
                "os_device_name": "", 
                "port_fcid": 69888, 
                "port_max_frame_size": 0, 
                "port_speed": 0, 
                "port_state": "online", 
                "port_supported_speeds": [], 
                "port_symbolic_name": "", 
                "port_type": null, 
                "port_wwn": "21:00:00:1b:32:17:63:91", 
                "serial_number": null, 
                "statistics": {
                    "dumped_frames": -1, 
                    "error_frames": -1, 
                    "fcp_control_requests": -1, 
                    "fcp_input_megabytes": -1, 
                    "fcp_input_requests": -1, 
                    "fcp_output_megabytes": -1, 
                    "fcp_output_requests": -1, 
                    "invalid_crc_count": -1, 
                    "invalid_tx_word_count": -1, 
                    "link_failure_count": -1, 
                    "lip_count": -1, 
                    "loss_of_signal_count": -1, 
                    "loss_of_sync_count": -1, 
                    "nos_count": -1, 
                    "primitive_seq_protocol_err_count": -1, 
                    "rx_frames": -1, 
                    "rx_words": -1, 
                    "seconds_since_last_reset": -1, 
                    "tx_frames": -1, 
                    "tx_words": -1
                }
            }
        ], 
        "driver_name": "ql2300.sys", 
        "driver_version": "9.1.8.6", 
        "fabric_name": "10:00:00:05:33:68:30:15", 
        "firmware_version": "4.06.01", 
        "hardware_version": "", 
        "manufacturer": "QLogic Corporation", 
        "model": "QLE2562", 
        "model_description": "QLogic QLE2562 Fibre Channel Adapter", 
        "node_wwn": "20:00:00:24:ff:2c:4e:52", 
        "option_rom_version": "2.16", 
        "os_device_name": "\\\\\\\\.\\\\Scsi5:", 
        "port_fcid": 70400, 
        "port_max_frame_size": 2048, 
        "port_speed": null, 
        "port_state": "online", 
        "port_supported_speeds": [
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4e:52", 
        "serial_number": "I07502", 
        "statistics": {
            "dumped_frames": -1, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": -1, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": -1, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 1, 
            "link_failure_count": 1, 
            "lip_count": -1, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": -1, 
            "primitive_seq_protocol_err_count": 0, 
            "rx_frames": -1, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": -1, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [], 
        "driver_name": "ql2300.sys", 
        "driver_version": "9.1.8.6", 
        "fabric_name": "00:00:00:00:00:00:00:00", 
        "firmware_version": "4.06.01", 
        "hardware_version": "", 
        "manufacturer": "QLogic Corporation", 
        "model": "QLE2562", 
        "model_description": "QLogic QLE2562 Fibre Channel Adapter", 
        "node_wwn": "20:00:00:24:ff:2c:4d:f3", 
        "option_rom_version": "2.16", 
        "os_device_name": "\\\\\\\\.\\\\Scsi4:", 
        "port_fcid": 0, 
        "port_max_frame_size": 2048, 
        "port_speed": 0, 
        "port_state": "link down", 
        "port_supported_speeds": [
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4d:f3", 
        "serial_number": "I48717", 
        "statistics": {
            "dumped_frames": -1, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": -1, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": -1, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 0, 
            "lip_count": -1, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": -1, 
            "primitive_seq_protocol_err_count": 0, 
            "rx_frames": -1, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": -1, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [], 
        "driver_name": "ql2300.sys", 
        "driver_version": "9.1.8.6", 
        "fabric_name": "00:00:00:00:00:00:00:00", 
        "firmware_version": "4.06.01", 
        "hardware_version": "", 
        "manufacturer": "QLogic Corporation", 
        "model": "QLE2562", 
        "model_description": "QLogic QLE2562 Fibre Channel Adapter", 
        "node_wwn": "20:00:00:24:ff:2c:4d:f2", 
        "option_rom_version": "2.16", 
        "os_device_name": "\\\\\\\\.\\\\Scsi3:", 
        "port_fcid": 0, 
        "port_max_frame_size": 2048, 
        "port_speed": 0, 
        "port_state": "link down", 
        "port_supported_speeds": [
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4d:f2", 
        "serial_number": "I48461", 
        "statistics": {
            "dumped_frames": -1, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": -1, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": -1, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 0, 
            "lip_count": -1, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": -1, 
            "primitive_seq_protocol_err_count": 0, 
            "rx_frames": -1, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": -1, 
            "tx_words": -1
        }
    }
]
""")
ports = [Port().update_not_none_values(item) for item in example]



__import__("pkg_resources").declare_namespace(__name__)

"""
This module fetch HBAAPI information from several sources, depending which ones are availble for the operating system.
Such sources are the actual hbaapi shared library, sysfs, etc...

The only entry points you should use are the PortsCollection class and the function get_port_collection.
Under the exmaples module, you can find some real-life examples of the outputs of this module.

The specific attirbutes available for each port are listed in FC_PORT_ATTRIBUTES
Their format or expected values are not documented here at this point. However, you can find this out by reading
the PortAssertions class in the tests modules
""" #pylint: disable-msg=W0105


from bunch import Bunch

WWN_OCTECT = '[0-9a-fA-F]{2}'
WWN_PATTERN = r'.*' + r'(?:0x)?' + r'[:-]?'.join(['(%s)' % WWN_OCTECT] * 8) + r'.*'

FC_PORT_ATTRIBUTES = ['node_wwn', 'port_wwn', 'port_fcid', 'port_state', 'port_type',
                      'port_symbolic_name', 'os_device_name', 'port_supported_speeds',
                      'port_speed', 'port_max_frame_size', 'fabric_name',
                      'driver_name', 'driver_version', 'manufacturer', 'serial_number',
                      'model', 'model_description', 'hardware_version', 'firmware_version',
                      'option_rom_version', 'discovered_ports', 'statistics', 'hct']
# TODO add tests for both sysfs and hbaapi implementations

FC_PORT_STATISTICS = ['dumped_frames', 'error_frames', 'fcp_control_requests', 'fcp_input_megabytes',
                      'fcp_input_requests', 'fcp_output_megabytes', 'fcp_output_requests', 'invalid_crc_count',
                      'invalid_tx_word_count', 'link_failure_count', 'lip_count', 'loss_of_signal_count',
                      'loss_of_sync_count', 'nos_count', 'primitive_seq_protocol_err_count',
                      'rx_frames', 'rx_words', 'seconds_since_last_reset', 'tx_frames', 'tx_words']

class Port(Bunch):
    """ Representation of a HBA port.
    For the list of available attributes, see the FC_PORT_ATTRIBUTES list
    """
    def __init__(self):
        Bunch.__init__(self)
        self._set_defaults()

    def _set_defaults(self):
        for attribute in FC_PORT_ATTRIBUTES:
            self.__setitem__(attribute, None) #pylint: disable-msg=E1101
        self.__setitem__('discovered_ports', list()) #pylint: disable-msg=E1101
        self.__setitem__('statistics', PortStatistics()) #pylint: disable-msg=E1101

    def update_not_none_values(self, other_port):
        for key, value in other_port.iteritems():
            our_value = self.get(key, None) #pylint: disable-msg=E1101
            if key == 'statistics':
                continue
            if our_value not in [-1, 0, None, '', [], ]:
                continue
            if value in [-1, 0, None, '', []]:
                continue
            self.__setitem__(key, value) #pylint: disable-msg=E1101
        if not other_port.get('statistics'): #pylint: disable-msg=E1101
            return self
        _ = self.statistics.update_not_none_values(other_port.get('statistics', dict())) #pylint: disable-msg=E1101
        return self

class PortStatistics(Port):
    def _set_defaults(self):
        for attribute in FC_PORT_STATISTICS:
            self.__setitem__(attribute, -1) #pylint: disable-msg=E1101

class PortsCollection(object):
    """ The entry point for this module. It provides a generator object trough iter_ports method, which yields Port
    objects, and the get_ports methods, which returns a list of all methods
    """
    def __init__(self):
        object.__init__(self)
        self._ports = {}

    def get_ports(self):
        return [port for port in self._ports.values()]

    def iter_ports(self):
        for port in self._ports.values():
            yield port

    def get_port_by_wwn(self, wwn):
        return self._ports[wwn]

    def add_port(self, port):
        if not self._ports.has_key(port.port_wwn):
            self._ports[port.port_wwn] = port
        else:
            self._ports[port.port_wwn].update_not_none_values(port)

    def remove_port_by_wwn(self, wwn):
        del(self._ports[wwn])

def get_ports_generator():
    """ Returns a composite component of Fiber Channel Ports generators that are available on this operating system
    """
    from .generators import get_list_of_generators, CompositeGenerator
    composite = CompositeGenerator()
    for leaf in get_list_of_generators():
        if not leaf.is_available():
            continue
        composite.add_generator(leaf)
    return composite

def get_ports_collection():
    """ Returns a unique collection of Fiber Channel Ports
    This is a simple wrapping usage of the PortsCollection case
    """
    collection = PortsCollection()
    generator = get_ports_generator()
    for port in generator.iter_ports():
        collection.add_port(port)
    return collection

__all__ = ['get_ports_collection', ]



__import__("pkg_resources").declare_namespace(__name__)

from contextlib import contextmanager
from bunch import Bunch
import ctypes
from .. import Generator
from ... import Port, PortStatistics, FC_PORT_STATISTICS

from glob import glob
from os.path import exists, join, sep, relpath

ROOT_FS = sep
OVERFLOW = '0xffffffffffffffff'
UNKNOWN_WWN = '0xffffffff'

def translate_stat_value_to_number(stat_value):
    return stat_value if isinstance(stat_value, int) else -1 if stat_value == OVERFLOW else int(stat_value, 16)

def translate_wwn(source_wwn):
    """ this function translates 0x010203040506070A to 01:02:03:04:05:06:07:0a
    """
    import re
    from ... import WWN_PATTERN
    from infi.dtypes.wwn import WWN
    if source_wwn == -1:
        return None
    return WWN(re.sub(WWN_PATTERN, r'\1:\2:\3:\4:\5:\6:\7:\8',
                  source_wwn if source_wwn != UNKNOWN_WWN else ':'.join(['ff'] * 8)).lower())

def translate_supported_speeds(source):
    """ this functions traslates ''1 Gbit, 2 Gbit, 4 Gbit, 8 Gbit' to [1,2,4,8]
    """
    if source == -1:
        return None
    return [int(item.replace(' Gbit', '')) for item in source.split(',')]

def translate_port_speed(source):
    if source == -1:
        return 0
    if source.lower() == 'unknown':
        return 0
    return int(source.replace(' Gbit', ''))

def translate_port_state(source):
    if source == -1:
        return None
    lower = source.lower()
    return None if lower in ['unknown', ] else lower

def translate_port_type(source):
    if source == -1:
        return None
    lower = source.lower().strip()
    return None if lower in ['unknown' , 'other', 'not present'] else lower

class Sysfs(Generator):
    def __init__(self):
        Generator.__init__(self)

    def _iter_fc_hosts(self):
        FC_HOST_BASEPATH = join(ROOT_FS, 'sys', 'class', 'fc_host')
        for path in glob(join(FC_HOST_BASEPATH, 'host*')):
            yield path, relpath(path, FC_HOST_BASEPATH).replace('host', '')

    def _iter_remote_fc_ports(self, fc_host_path):
        for rport in glob(join(fc_host_path, 'device', 'rport*')):
            if exists(join(rport, 'fc_remote_ports')):
                # this is how its on ubuntu
                for path in glob(join(rport, 'fc_remote_ports', '*')):
                    yield path
            else:
                for path in glob(join(rport, 'fc_remote_ports*')):
                    yield path

    def get_file_content(self, filepath):
        # The purpose of this method is to return the file's content with no exceptions.
        # If we fail to collect the information, just return None (that's OK)
        #
        # If a wildcard '*' is in filepath, we will glob for the list of matching files, and choose the first one
        # This is done to pretect ourself from cases such as model_description is just named model_desc
        if '*' in filepath:
            filelist = glob(filepath)
            if not filelist:
                return None
            filepath = filelist[0]
        try:
            with open(filepath) as fd:
                return fd.read().strip('\n').strip()
        except IOError:
            return -1

    def _populate_port_attributes_from_fc_host(self, port, base_path):
        port.port_fcid = self.get_file_content(join(base_path, 'port_id'))
        port.port_wwn = translate_wwn(self.get_file_content(join(base_path, 'port_name')))
        port.node_wwn = translate_wwn(self.get_file_content(join(base_path, 'node_name')))
        port.port_state = translate_port_state(self.get_file_content(join(base_path, 'port_state')))
        port.port_speed = translate_port_speed(self.get_file_content(join(base_path, 'speed')))
        port.port_type = translate_port_type(self.get_file_content(join(base_path, 'port_type')))
        port.port_supported_speeds = translate_supported_speeds(
                                                self.get_file_content(join(base_path, 'supported_speeds')))
        port.fabric_name = translate_wwn(self.get_file_content(join(base_path, 'fabric_name')))
        port.port_symbolic_name = self.get_file_content(join(base_path, 'symbolic_name'))
	self._populate_local_port_hct(port, base_path)

    def _populate_port_attributes_from_scsi_host(self, port, base_path):
        port.model = self.get_file_content(join(base_path, 'model*name'))
        port.model_description = self.get_file_content(join(base_path, 'model*desc*'))
        port.driver_version = self.get_file_content(join(base_path, '*dr*v*_version'))
        port.serial_number = self.get_file_content(join(base_path, 'serial*num'))
        port.hardware_version = self.get_file_content(join(base_path, 'fw*e*'))
        port.option_rom_version = self.get_file_content(join(base_path, 'optrom_*_version'))

    def _populate_remote_port_hct(self, port, base_path, local_port):
        from re import compile
        pattern = compile(r"(?P<host>\d+)[^\d](?P<channel>\d+)[^\d](?P<target>\d)")
        result = pattern.search(base_path).groupdict()
        # TODO get the target number from the scsi_target_id_file inside base_path
        target_path = join(base_path, "scsi_target_id")
        target_id = open(target_path).read().strip("\n").strip() if exists(target_path) else "-1"
        port.hct = (local_port.hct[0], int(result['channel']), int(target_id))

    def _populate_local_port_hct(self, port, base_path):
        from re import compile
        pattern = compile(r"host(?P<host>\d+)")
        result = pattern.search(base_path).groupdict()
        port.hct = (int(result['host']), -1, -1)

    def _populate_remote_port_attributes_from_fc_host(self, port, base_path, local_port):
        self._populate_port_attributes_from_fc_host(port, base_path)
        self._populate_remote_port_hct(port, base_path, local_port)

    def _populate_port_statistics_from_fc_host(self, port, base_path):
        # currently, FC_PORT_STATISTICS match the files in sysfs, so its easy to this
        stats = PortStatistics()
        for stat_name in FC_PORT_STATISTICS:
            stat_value = self.get_file_content(join(base_path, 'statistics', stat_name))
            # only one name is fucked up here
            setattr(stats, stat_name.replace('prim', 'primitive'), translate_stat_value_to_number(stat_value))
        port.statistics = stats

    def _populate_discovered_ports(self, port, fc_host_path):
        port.discovered_ports = []
        for remote_fc_port_path in self._iter_remote_fc_ports(fc_host_path):
            remote_port = Port()
            self._populate_remote_port_attributes_from_fc_host(remote_port, remote_fc_port_path, port)
            if remote_port.hct[2] == -1:
                # this is not a real target, just a switch port
                continue
            if remote_port.port_state == "not present":
                # this port is no longer connected
                continue
            port.discovered_ports.append(remote_port)

    def iter_ports(self):
        SCSI_HOST_BASEPATH = join(ROOT_FS, 'sys', 'class', 'scsi_host')
        for fc_host_path, host_id in self._iter_fc_hosts():
            # fc_host_path is full path
            scsi_host_path = join(SCSI_HOST_BASEPATH, 'host%s' % host_id)
            port = Port()
            self._populate_port_attributes_from_fc_host(port, fc_host_path)
            self._populate_port_attributes_from_scsi_host(port, scsi_host_path)
            self._populate_port_statistics_from_fc_host(port, fc_host_path)
            self._populate_discovered_ports(port, fc_host_path)
            yield port

    @classmethod
    def is_available(cls):
        FC_HOST_BASEPATH = join(ROOT_FS, 'sys', 'class', 'fc_host')
        return exists(FC_HOST_BASEPATH)




import unittest
import mock
import logging
logger = logging.getLogger(__name__)

#pylint: disable-all

class MockGenerator(object):
    def __init__(self):
        object.__init__(self)

    def iter_ports(self):
        return []

    @classmethod
    def is_available(cls):
        return True

class TestCase(unittest.TestCase):
    def test_imports(self):
        from .. import get_ports_collection, get_ports_generator
        from ..generators import Generator, CompositeGenerator
        from .. import Port, PortsCollection
    pass

    @mock.patch("infi.hbaapi.generators.get_list_of_generators")
    def test_iterface(self, mock):
        from ..generators.tests import get_list_of_generators
        from .. import get_ports_collection, get_ports_generator
        mock.return_value = get_list_of_generators()
        generator = get_ports_generator()
        self.assertEquals(mock.call_count, 1)
        available_generators = [item for item in generator.iter_generators()]
        self.assertEquals(len(available_generators), 1)
        collection = get_ports_collection()
        self.assertEquals(len(collection.get_ports()), 0)

    def test_real_thing(self):
        from .. import get_ports_collection
        ports = [port for port in get_ports_collection().iter_ports()]

    @mock.patch("infi.hbaapi.generators.get_list_of_generators")
    def test_with_mock_generator(self, mock):
        mock.return_value = [MockGenerator]
        from .. import get_ports_collection
        ports = [port for port in get_ports_collection().iter_ports()]

class PortAssertions(object):
    def __init__(self, test_case):
        self.test_case = test_case

    def _assert_statistics(self, port):
        from infi.hbaapi import FC_PORT_STATISTICS
        for name in FC_PORT_STATISTICS:
            value = getattr(port.statistics, name)
            self.test_case.assertIsInstance(value, int, "%s is not an instance of int" % repr(value))

    def _assert_wwn(self, wwn):
        import re
        from infi.dtypes.wwn import WWN
        from .. import WWN_PATTERN
        logger.debug(wwn)
        self.test_case.assertIsNotNone(wwn)
        self.test_case.assertIsInstance(wwn, WWN)

    def _assert_supported_speeds(self, port):
        self.test_case.assertIsInstance(port.port_supported_speeds, list)
        for item in port.port_supported_speeds:
            self.test_case.assertIsInstance(item, int)

    def _assert_port_speed(self, port_speed):
        self.test_case.assertIn(port_speed, [0, 1, 2, 4, 8])

    def _assert_port_state(self, port_state):
        self.test_case.assertIn(port_state, [None, 'online', 'offline', 'bypassed', 'diagnostics',
                                   'link down', 'error', 'loopback'])

    def _assert_port_type(self, port_type):
        self.test_case.assertIn(port_type, [None, 'N', 'NL', 'FL', 'F', 'E', 'G', 'L', 'PTP'])

    def _assert_discovered_ports(self, discovered_ports):
        self.test_case.assertIsInstance(discovered_ports, list)
        for port in discovered_ports:
            self.assert_port(port, True)

    def assert_port(self, port, remote_port=False):
        self._assert_wwn(port.port_wwn)
        self._assert_wwn(port.node_wwn)
        self._assert_wwn(port.fabric_name)
        self._assert_supported_speeds(port)
        self._assert_port_speed(port.port_speed)
        self._assert_port_state(port.port_state)
        self._assert_port_type(port.port_type)
        if not remote_port:
            self._assert_statistics(port)




import unittest

class RemoteHCTLTestCase(unittest.TestCase):
    def test_remote_port_hctl(self):
        from infi.hbaapi import get_ports_collection
        local_ports = get_ports_collection().get_ports()
        for local_port in local_ports:
            remote_ports = local_port.discovered_ports
            for remote_port in remote_ports:
                self.assertFalse(remote_port.hct is None)




__import__("pkg_resources").declare_namespace(__name__)

#pylint: disable-all

import logging
import unittest
import mock
from contextlib import contextmanager, nested
from os.path import exists, join, sep, dirname, pardir, abspath
from ... import hbaapi
from .. import headers, c_api

logger = logging.getLogger(__name__)

def _translate_wwn(source):
    return ''.join([chr(item) for item in source])

ADAPTER_NAMES = ['lsiSAS', 'qlogic1', 'qlogic2']
ADAPTER_HANDLES = {'lsiSAS': 1, 'qlogic1': 2, 'qlogic2': 3}
ADAPTER_ATTRIBUTES_BY_HANDLE = {1: NotImplementedError,
                                2: dict(
    Manufacturer='QLogic Corporation',
    SerialNumber='I48461',
    Model='QLE2562',
    ModelDescription='QLogic QLE2562 Fibre Channel Adapter',
    NodeWWN=_translate_wwn([32, 0, 0, 36, 255, 44, 77, 242, ]),
    NodeSymbolicName='QLE2562 FW:v4.06.01 DVR:v9.1.8.6',
    HardwareVersion='',
    DriverVersion='9.1.8.6',
    OptionROMVersion='2.16',
    FirmwareVersion='4.06.01',
    VendorSpecificID=624038007,
    NumberOfPorts=1,
    DriverName='ql2300.sys',),
                                3: dict(
    Manufacturer='QLogic Corporation',
    SerialNumber='I48717',
    Model='QLE2562',
    ModelDescription='QLogic QLE2562 Fibre Channel Adapter',
    NodeWWN=_translate_wwn([32, 0, 0, 36, 255, 44, 77, 243, ]),
    NodeSymbolicName='QLE2562 FW:v4.06.01 DVR:v9.1.8.6',
    HardwareVersion='',
    DriverVersion='9.1.8.6',
    OptionROMVersion='2.16',
    FirmwareVersion='4.06.01',
    VendorSpecificID=624038007,
    NumberOfPorts=1,
    DriverName='ql2300.sys',
)}

PORT_ATTRIBUTES_BY_ADAPTER_HANDLE = {2:[dict(
    NodeWWN=_translate_wwn([32, 0, 0, 36, 255, 44, 77, 242, ]),
    PortWWN=_translate_wwn([33, 0, 0, 36, 255, 44, 77, 242, ]),
    PortFcId=0,
    PortType=1,
    PortState=6,
    PortSupportedClassofService=8,
    PortSupportedFc4Types=[0] * 32,
    PortActiveFc4Types=[0] * 32,
    PortSymbolicName='',
    OSDeviceName='\\\\.\\Scsi3:',
    PortSuggestedSpeed=26,
    PortSpeed=0,
    PortMaxFrameSize=2048,
    FabricName=_translate_wwn([0, 0, 0, 0, 0, 0, 0, 0]),
    NumberOfDiscoveredPorts=1,),
                                        ],
                                     3:[dict(
    NodeWWN=_translate_wwn([32, 0, 0, 36, 255, 44, 77, 243, ]),
    PortWWN=_translate_wwn([33, 0, 0, 36, 255, 44, 77, 243, ]),
    PortFcId=0,
    PortType=1,
    PortState=6,
    PortSupportedClassofService=8,
    PortSupportedFc4Types=[0] * 32,
    PortActiveFc4Types=[0] * 32,
    PortSymbolicName='',
    OSDeviceName='\\\\.\\Scsi3:',
    PortSuggestedSpeed=26,
    PortSpeed=0,
    PortMaxFrameSize=2048,
    FabricName=_translate_wwn([0, 0, 0, 0, 0, 0, 0, 0]),
    NumberOfDiscoveredPorts=1,),
                                        ]}

REMOTE_PORT_ATTRIBUTES = {
    2: [[dict(
    NodeWWN=_translate_wwn([1, 2, 3, 4, 5, 6, 7, 8]),
    PortWWN=_translate_wwn([1, 2, 3, 4, 5, 6, 7, 8]),
    PortFcId=123,
    PortType=0,
    PortState=1,
    PortSupportedClassofService=0,
    PortSupportedFc4Types=[0] * 32,
    PortActiveFc4Types=[0] * 32,
    PortSymbolicName='',
    OSDeviceName='',
    PortSuggestedSpeed=0,
    PortSpeed=0,
    PortMaxFrameSize=0,
    FabricName=_translate_wwn([0, 0, 0, 0, 0, 0, 0, 0]),
    NumberOfDiscoveredPorts=0,)]],
    3: [[dict(
    NodeWWN=_translate_wwn([1, 2, 3, 4, 5, 6, 7, 8]),
    PortWWN=_translate_wwn([1, 2, 3, 4, 5, 6, 7, 8]),
    PortFcId=123,
    PortType=0,
    PortState=1,
    PortSupportedClassofService=0,
    PortSupportedFc4Types=[0] * 32,
    PortActiveFc4Types=[0] * 32,
    PortSymbolicName='',
    OSDeviceName='',
    PortSuggestedSpeed=0,
    PortSpeed=0,
    PortMaxFrameSize=0,
    FabricName=_translate_wwn([0, 0, 0, 0, 0, 0, 0, 0]),
    NumberOfDiscoveredPorts=0,)], ]}

PORT_STATISTICS_BY_ADAPTER_HANDLE = {2:[dict(
    SecondsSinceLastReset= -1,
    TxFrames= -1,
    TxWords= -1,
    RxFrames= -1,
    RxWords= -1,
    LIPCount= -1,
    NOSCount= -1,
    ErrorFrames= -1,
    DumpedFrames= -1,
    LinkFailureCount=0,
    LossOfSyncCount=0,
    LossOfSignalCount=0,
    PrimitiveSeqProtocolErrCount=0,
    InvalidTxWordCount=0,
    InvalidCRCCount=0,
)],
                                     3:[dict(
    SecondsSinceLastReset= -1,
    TxFrames= -1,
    TxWords= -1,
    RxFrames= -1,
    RxWords= -1,
    LIPCount= -1,
    NOSCount= -1,
    ErrorFrames= -1,
    DumpedFrames= -1,
    LinkFailureCount=0,
    LossOfSyncCount=0,
    LossOfSignalCount=0,
    PrimitiveSeqProtocolErrCount=0,
    InvalidTxWordCount=0,
    InvalidCRCCount=0,
)]}

class GeneratorTestCase(unittest.TestCase):
    pass

    @mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetVersion")
    def test_get_version(self, api_mock):
        from .. import c_api
        api_mock.return_value = headers.HBA_UINT32(2)
        self.assertEquals(api_mock.return_value, c_api.HBA_GetVersion())

    @contextmanager
    def _mock_open_close_library(self):
        with nested(mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_LoadLibrary"),
                               mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_FreeLibrary")
                               ) as (load, free):
            load.return_value = 0
            free.return_value = 0
            yield (load, free)

    @contextmanager
    def _mock_get_number_of_adapters(self, expected=2):
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetNumberOfAdapters") as api_mock:
            api_mock.return_value = expected
            yield api_mock

    @contextmanager
    def _mock_get_adapter_name(self):
        def side_effect(*args, **kwargs):
            adapter_index = args[0]
            adapter_name_ctypes_object = args[1]
            if adapter_index > len(ADAPTER_NAMES) - 1:
                raise c_api.InconsistencyError
            adapter_name_ctypes_object.value = ADAPTER_NAMES[adapter_index]
            return 0
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetAdapterName") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_open_close_adapter(self):
        def side_effect(*args, **kwargs):
            adapter_name_ctypes_object = args[0]
            return headers.HBA_HANDLE(ADAPTER_HANDLES.get(adapter_name_ctypes_object.value))
        with nested(mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_OpenAdapter"),
                    mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_CloseAdapter"),
                    ) as (open, close):
            open.side_effect = side_effect
            close.return_value = 0
            yield (open, close)

    @contextmanager
    def _mock_get_adapter_attributes(self):
        def side_effect(*args, **kwargs):
            adapter_handle, buffer = args
            attributes = ADAPTER_ATTRIBUTES_BY_HANDLE[adapter_handle.value]
            if attributes is NotImplementedError:
                raise NotImplementedError
            instance = headers.HBA_AdapterAttributes()
            for key, value in attributes.items():
                setattr(instance, key, value)
            buffer.raw = headers.HBA_AdapterAttributes.write_to_string(instance)
            return 0
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetAdapterAttributes") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_get_port_attributes(self):
        def side_effect(*args, **kwargs):
            adapter_handle, port_index, buffer = args
            attributes = PORT_ATTRIBUTES_BY_ADAPTER_HANDLE[adapter_handle.value][port_index]
            if attributes is NotImplementedError:
                raise NotImplementedError
            instance = headers.HBA_PortAttributes(**attributes)
            buffer.raw = headers.HBA_PortAttributes.write_to_string(instance)
            return 0
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetAdapterPortAttributes") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_get_remote_port_attributes(self):
        def side_effect(*args, **kwargs):
            adapter_handle, port_index, remote_port_index, buffer = args
            attributes = REMOTE_PORT_ATTRIBUTES[adapter_handle.value][port_index][remote_port_index]
            if attributes is NotImplementedError:
                raise NotImplementedError
            instance = headers.HBA_PortAttributes(**attributes)
            buffer.raw = headers.HBA_PortAttributes.write_to_string(instance)
            return 0
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetDiscoveredPortAttributes") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_get_port_statistics(self):
        def side_effect(*args, **kwargs):
            adapter_handle, port_index, buffer = args
            attributes = PORT_STATISTICS_BY_ADAPTER_HANDLE[adapter_handle.value][port_index]
            if attributes is NotImplementedError:
                raise NotImplementedError
            instance = headers.HBA_PortStatistics()
            for key, value in attributes.items():
                setattr(instance, key, value)
            buffer.raw = headers.HBA_PortStatistics.write_to_string(instance)
            return 0
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetPortStatistics") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_get_fcp4_statistics(self):
        def side_effect(*args, **kwargs):
            from ..headers import HBA_STATUS_ERROR_UNSUPPORTED_FC4
            return HBA_STATUS_ERROR_UNSUPPORTED_FC4
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetFC4Statistics") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    @contextmanager
    def _mock_get_fcp_target_mappings(self):
        def side_effect(*args, **kwargs):
            from ..headers import HBA_STATUS_ERROR_NOT_SUPPORTED
            return HBA_STATUS_ERROR_NOT_SUPPORTED
        with mock.patch("infi.hbaapi.generators.hbaapi.c_api.HBA_GetFcpTargetMappingV2") as api_mock:
            api_mock.side_effect = side_effect
            yield api_mock

    def test_get_number_of_adapters(self):
        from .. import c_api
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters()
                    ) as (_, number_of_adapters):
            self.assertEquals(number_of_adapters.return_value, c_api.HBA_GetNumberOfAdapters())

    def test_get_port_adapter_raises_exception(self):
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters(1),
                    self._mock_get_adapter_name(),
                    self._mock_open_close_adapter(),
                    self._mock_get_adapter_attributes(),
                    ) as (_, _, _, _, api_mock):
            from .. import HbaApi
            ports = [port for port in HbaApi().iter_ports()]
            self.assertEquals(1, api_mock.call_count)

    def test_mock(self):
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters(3),
                    self._mock_get_adapter_name(),
                    self._mock_open_close_adapter(),
                    self._mock_get_adapter_attributes(),
                    self._mock_get_port_attributes(),
                    self._mock_get_port_statistics(),
                    self._mock_get_fcp4_statistics(),
                    self._mock_get_remote_port_attributes(),
                    self._mock_get_fcp_target_mappings(),
                    ):
            ports = self._get_ports()
            from infi.hbaapi.tests import PortAssertions
            port_test_class = PortAssertions(self)
            for port in ports:
                port_test_class.assert_port(port)
                self.assertEquals(port.port_speed, 0)
                self.assertEquals(port.port_supported_speeds, [2, 4, 8])

    def _get_ports(self):
        from .. import HbaApi
        return [port for port in HbaApi().iter_ports()]

    def test_for_inconsistency__adapter_index_invalid_at_adapter_name(self):
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters(5),
                    self._mock_get_adapter_name(),
                    self._mock_open_close_adapter(),
                    self._mock_get_adapter_attributes(),
                    self._mock_get_port_attributes(),
                    self._mock_get_port_statistics(),
                    self._mock_get_fcp4_statistics(),
                    self._mock_get_remote_port_attributes(),
                    self._mock_get_fcp_target_mappings()
                    ):
            self.assertRaises(c_api.InconsistencyError, self._get_ports)

    def test_for_inconsistency__adapter_index_invalid_at_adapter_attributes(self):
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters(1),
                    self._mock_get_adapter_name(),
                    self._mock_open_close_adapter(),
                    self._mock_get_adapter_attributes(),
                    self._mock_get_port_attributes(),
                    self._mock_get_port_statistics(),
                    self._mock_get_fcp4_statistics(),
                    ) as (_, _, _, _, mock_api, _, _, _):
            mock_api.side_effect = c_api.InconsistencyError
            self.assertRaises(c_api.InconsistencyError, self._get_ports)

    def test_for_inconsistency__remote_port_has_gone_away(self):
        with nested(
                    self._mock_open_close_library(),
                    self._mock_get_number_of_adapters(3),
                    self._mock_get_adapter_name(),
                    self._mock_open_close_adapter(),
                    self._mock_get_adapter_attributes(),
                    self._mock_get_port_attributes(),
                    self._mock_get_port_statistics(),
                    self._mock_get_fcp4_statistics(),
                    self._mock_get_remote_port_attributes()
                    ) as (_, _, _, _, _, _, _, _, mock_api):
            mock_api.side_effect = c_api.InconsistencyError
            self.assertRaises(c_api.InconsistencyError, self._get_ports)



__import__("pkg_resources").declare_namespace(__name__)

import logging
from contextlib import contextmanager
from bunch import Bunch
import ctypes
from .. import Generator
from ... import Port, PortStatistics, FC_PORT_STATISTICS
from infi.dtypes.wwn import WWN
import c_api, headers
import binascii

log = logging.getLogger(__name__)

WELL_KNOWN_FC_ADDRESSES = [
    0xFFFFF5, # Multicast server
    0xFFFFF6, # Clock Sync server
    0xFFFFF7, # KDC (key distribution)
    0xFFFFF8, # Alias server (for multicast, or hunt groups)
    0xFFFFF9, # QoS information
    0xFFFFFA, # Management server
    0xFFFFFB, # Time server
    0xFFFFFC, # Directory server
    0xFFFFFD, # Fabric Controller
    0xFFFFFE, # Fabric Login server
]

class HbaApi(Generator):
    def __init__(self):
        Generator.__init__(self)

    def _iter_adapters(self):
        for adapter_index in range (0, c_api.HBA_GetNumberOfAdapters()):
            adapter_name = ctypes.create_string_buffer(headers.MAX_ADAPTERNAME_LENGTH)
            c_api.HBA_GetAdapterName(adapter_index, adapter_name)
            log.debug("yield adapter index {} name {!r}".format(adapter_index, adapter_name.value))
            yield adapter_name

    def _get_adapter_attributes(self, adapter_handle):
        buff = ctypes.c_buffer(headers.HBA_AdapterAttributes.min_max_sizeof().max) #pylint: disable-msg=W0622,E1101
        c_api.HBA_GetAdapterAttributes(adapter_handle, buff)
        adapter_attributes = headers.HBA_AdapterAttributes.create_from_string(buff) #pylint: disable-msg=E1101
        c_api.HBA_GetAdapterAttributes(adapter_handle, buff)
        return adapter_attributes

    def _get_port_attributes(self, adapter_handle, port_index):
        buffer = ctypes.c_buffer(headers.HBA_PortAttributes.min_max_sizeof().max) #pylint: disable-msg=W0622,E1101
        c_api.HBA_GetAdapterPortAttributes(adapter_handle, port_index, buffer)
        port_attributes = headers.HBA_PortAttributes.create_from_string(buffer) #pylint: disable-msg=E1101
        return port_attributes

    def _get_remote_port_attributes(self, adapter_handle, port_index, remote_port_index):
        buff = ctypes.c_buffer(headers.HBA_PortAttributes.min_max_sizeof().max) #pylint: disable-msg=W0622,E1101
        c_api.HBA_GetDiscoveredPortAttributes(adapter_handle, port_index, remote_port_index, buff)
        remote_port_attributes = headers.HBA_PortAttributes.create_from_string(buff) #pylint: disable-msg=E1101
        return remote_port_attributes

    def _get_remote_ports(self, adapter_handle, port_index, number_of_remote_ports):
        remote_ports = []
        for remote_port_index in range(0, number_of_remote_ports):
            remote_port_attributes = self._get_remote_port_attributes(adapter_handle, port_index, remote_port_index)
            remote_port = get_port_object(None, remote_port_attributes)
            log.debug("Found remote port {!r}".format(remote_port.port_wwn))
            if remote_port.port_fcid in WELL_KNOWN_FC_ADDRESSES:
                msg = "port {!r} is a well-known FC address with fcid {!r}"
                log.debug(msg.format(remote_port.port_wwn, remote_port.port_fcid))
            elif remote_port.port_state == 'offline':
                msg = "remote port {!r} is offline"
                log.debug(msg.format(remote_port.port_wwn))
            else:
                remote_ports.append(remote_port)
        return remote_ports

    def _populate_local_port_hct(self, port):
        from re import compile
        pattern = compile(r"(?P<host>\d+)$")
        result = pattern.search(port.os_device_name.strip(':')).groupdict()
        port.hct = (int(result['host']), -1, -1)

    def _get_local_port(self, adapter_handle, adapter_attributes, port_index):
        port_attributes = self._get_port_attributes(adapter_handle, port_index)
        port = get_port_object(adapter_attributes, port_attributes)
        logging.debug("checking local fc port {!r}".format(port.port_wwn))
        self._populate_local_port_hct(port)
        wwn_buffer = self._extract_wwn_buffer_from_port_attributes(port_attributes)
        number_of_remote_ports = port_attributes.NumberOfDiscoveredPorts
        remote_ports = self._get_remote_ports(adapter_handle, port_index, number_of_remote_ports)
        port.discovered_ports = remote_ports
        port_statistics = self._get_local_port_statistics(adapter_handle, port_index, wwn_buffer)
        port.statistics = port_statistics
        port_mappings = self._get_local_port_mappings(adapter_handle, wwn_buffer)
        for remote_port in port.discovered_ports:
            channel, target = port_mappings.get(remote_port.port_wwn, (-1, -1))
            remote_port.hct = (port.hct[0], channel, target)
        return port

    def _get_local_port_statistics(self, adapter_handle, port_index, wwn_buffer):
        port_statistics = PortStatistics()
        hba_port_stats, hba_fc4_stats = None, None

        buffer = ctypes.c_buffer(headers.HBA_PortStatistics.min_max_sizeof().max) #pylint: disable-msg=W0622,E1101
        try:
            c_api.HBA_GetPortStatistics(adapter_handle, port_index, buffer)
        except RuntimeError:
            # Default QLogic drivers on Windows don't support this function - return defaults
            return port_statistics
        hba_port_stats = headers.HBA_PortStatistics.create_from_string(buffer) #pylint: disable-msg=E1101

        try:
            buffer = ctypes.c_buffer(headers.HBA_FC4Statistics.min_max_sizeof().max) #pylint: disable-msg=W0622,E1101
            c_api.HBA_GetFC4Statistics(adapter_handle, wwn_buffer, 2, buffer)
            hba_fc4_stats = headers.HBA_FC4Statistics.create_from_string(buffer) #pylint: disable-msg=E1101
        except NotImplementedError:
            pass
        except RuntimeError, exception:
            return_code = exception.args[0]
            if return_code in [headers.HBA_STATUS_ERROR_UNSUPPORTED_FC4, headers.HBA_STATUS_ERROR_ILLEGAL_WWN]:
                pass
            else:
                raise

        self._merge_statistics(port_statistics, hba_port_stats, hba_fc4_stats)
        return port_statistics

    def _merge_statistics(self, port_statistics, hba_port_stats=None, hba_fc4_stats=None):
        # FC_PORT_STATISTICS and HBA_PortStatistics and HBA_FC4Statistics are very similar
        def _merge_hba_port_stat(name):
            # some names are in different formatting
            alternative_name = name.replace('nos', 'NOS').replace('lip', 'LIP').replace('crc', 'CRC')
            hbaapi_name = ''.join([(item if item.isupper() else item.capitalize()) \
                                   for item in alternative_name.split('_')])
            setattr(port_statistics, name, -1 if hba_port_stats is None else getattr(hba_port_stats, hbaapi_name))

        def _merge_hba_fcp4_stat(name):
            hbaapi_name = ''.join([item.capitalize() for item in name.replace('fcp_', '').split('_')])
            setattr(port_statistics, name, -1 if hba_fc4_stats is None else getattr(hba_fc4_stats, hbaapi_name))

        for name in FC_PORT_STATISTICS:
            if name.startswith('fcp_'):
                _merge_hba_fcp4_stat(name)
            else:
                _merge_hba_port_stat(name)

    def _mappings_to_dict(self, mappings):
        result = dict()
        for entry in mappings.entry:
            if entry.FcId.PortWWN == '':
                log.debug("Found an empty mapping")
                continue
            log.debug("Found mapping {!r} to {!r}".format(entry.ScsiId, binascii.hexlify(entry.FcId.PortWWN)))
            key = translate_wwn(entry.FcId.PortWWN)
            value = (entry.ScsiId.ScsiBusNumber, entry.ScsiId.ScsiTargetNumber)
            result[key] = value
        log.debug("{!r}".format(result))
        return result

    def _extract_wwn_buffer_from_port_attributes(self, port_attributes):
        return ctypes.c_uint64(headers.UNInt64.create_from_string(port_attributes.PortWWN))

    def _build_empty_mappings_struct(self, number_of_elements):
        element_size = headers.HBA_FcpScsiEntryV2.min_max_sizeof().max
        entry = headers.HBA_FcpScsiEntryV2.create_from_string('\x00'*element_size)
        struct = headers.HBA_FCPTargetMappingV2(NumberOfEntries=number_of_elements,
                                                entry=[entry for index in range(number_of_elements)])
        return struct

    def _get_local_port_mappings_count(self, adapter_handle, wwn_buffer):
        struct = self._build_empty_mappings_struct(0)
        size = struct.sizeof(struct)
        mappings_buffer = ctypes.c_buffer('\x00'*size, size)
        try:
            c_api.HBA_GetFcpTargetMappingV2(adapter_handle, wwn_buffer, mappings_buffer)
        except RuntimeError, exception:
            return_code = exception.args[0]
            if return_code == headers.HBA_STATUS_ERROR_MORE_DATA:
                return headers.UNInt32.create_from_string(mappings_buffer)
        return 0

    def _get_local_port_mappings(self, adapter_handle, wwn_buffer):
        try:
            number_of_entries = self._get_local_port_mappings_count(adapter_handle, wwn_buffer)
            struct = self._build_empty_mappings_struct(number_of_entries)
            size = struct.sizeof(struct)
            mappings_buffer = ctypes.c_buffer(struct.write_to_string(struct), size)
            c_api.HBA_GetFcpTargetMappingV2(adapter_handle, wwn_buffer, mappings_buffer)
        except RuntimeError, exception:
            msg ="failed to fetch port mappings for local wwn {!r}"
            log.debug(msg.format(binascii.hexlify(wwn_buffer.raw)))
            return_code = exception.args[0]
            if return_code == headers.HBA_STATUS_ERROR_ILLEGAL_WWN:
                log.debug("error code is HBA_STATUS_ERROR_ILLEGAL_WWN")
                return {}
            if return_code == headers.HBA_STATUS_ERROR_NOT_SUPPORTED:
                log.debug("error code is HBA_STATUS_ERROR_NOT_SUPPORTED")
                return {}
            if return_code == headers.HBA_STATUS_ERROR_MORE_DATA:
                return self._get_local_port_mappings(adapter_handle, wwn_buffer,
                                                     buffer_size*2)
            else:
                raise
        
        mappings = headers.HBA_FCPTargetMappingV2.create_from_string(mappings_buffer)
        return self._mappings_to_dict(mappings)

    def iter_ports(self):
        with self.hbaapi_library():
            for adapter_name in self._iter_adapters():
                with self.hbaapi_adapter(adapter_name) as adapter_handle:
                    try:
                        adapter_attributes = self._get_adapter_attributes(adapter_handle)
                    except NotImplementedError:
                        # because of weird HBAs, like VMWare's "LSI Adapter, SAS 3000 Series"
                        logging.debug("found a weird host bus adapter, named: %s", adapter_name.value)
                        continue
                    for port_index in range(0, adapter_attributes.NumberOfPorts):
                        log.debug("yield port index {} for adapter name {}".format(port_index, adapter_name))
                        local_port = self._get_local_port(adapter_handle, adapter_attributes, port_index)
                        yield local_port

    @classmethod
    def is_available(cls):
        try:
            c_api.HBA_GetVersion._get_function() #pylint: disable-msg=W0212,E1101 
            return True
        except OSError:
            pass
        return False

    @contextmanager
    def hbaapi_library(self):
        c_api.HBA_LoadLibrary()
        try:
            yield
        finally:
            c_api.HBA_FreeLibrary()

    @contextmanager
    def hbaapi_adapter(self, adapter_name):
        handle = c_api.HBA_OpenAdapter(adapter_name)
        try:
            yield handle
        finally:
            c_api.HBA_CloseAdapter(handle)

def translate_wwn(source):
    return WWN(binascii.hexlify(source if source != '' else '\x00'*8))

def translate_port_type(number):
    return headers.HBA_PORTTYPE[str(number)]

def translate_port_speed(source):
    """ PortSpeed indicates the signalling bit rate at which this port is currently operating.
    """
    if source in [headers.HBA_PORTSPEED_UNKNOWN, headers.HBA_PORTSPEED_NOT_NEGOTIATED]:
        return 0
    translation = {headers.HBA_PORTSPEED_1GBIT: 1,
                   headers.HBA_PORTSPEED_2GBIT: 2,
                   headers.HBA_PORTSPEED_10GBIT: 10,
                   headers.HBA_PORTSPEED_4GBIT: 4,
                   headers.HBA_PORTSPEED_8GBIT: 8,
                   }
    return translation[source]

def translate_port_supported_speeds(source):
    """ PortSupportedSpeed indicates the signalling bit rates at which this port may operate.
    """
    result = []
    translation = {headers.HBA_PORTSPEED_1GBIT: 1,
                   headers.HBA_PORTSPEED_2GBIT: 2,
                   headers.HBA_PORTSPEED_10GBIT: 10,
                   headers.HBA_PORTSPEED_4GBIT: 4,
                   headers.HBA_PORTSPEED_8GBIT: 8,
                   }
    keys = translation.keys()
    keys.sort(reverse=True)
    for key in keys:
        if key & source:
            result.append(translation[key])
            source -= key
    result.sort()
    return result

def translate_port_state(source):
    translation = {headers.HBA_PORTSTATE_UNKNOWN: None,
                   headers.HBA_PORTSTATE_ONLINE: 'online',
                   headers.HBA_PORTSTATE_OFFLINE: 'offline',
                   headers.HBA_PORTSTATE_BYPASSED: 'bypassed',
                   headers.HBA_PORTSTATE_DIAGNOSTICS: 'diagnostics',
                   headers.HBA_PORTSTATE_LINKDOWN: 'link down',
                   headers.HBA_PORTSTATE_ERROR: 'error',
                   headers.HBA_PORTSTATE_LOOPBACK: 'lookback'}
    return translation[source]

def get_port_object(adapter_attributes=Bunch(), port_attributes=Bunch()):
    kwargs = {}
    kwargs['node_wwn'] = translate_wwn(getattr(adapter_attributes, "NodeWWN", None) or \
                                       getattr(port_attributes, "NodeWWN", None))
    kwargs['port_wwn'] = translate_wwn(getattr(port_attributes, "PortWWN"))
    kwargs['port_fcid'] = getattr(port_attributes, "PortFcId")
    kwargs['port_type'] = translate_port_type(getattr(port_attributes, "PortType"))
    kwargs['port_state'] = translate_port_state(getattr(port_attributes, "PortState"))
    kwargs['port_speed'] = translate_port_speed(getattr(port_attributes, "PortSpeed"))
    kwargs['port_symbolic_name'] = getattr(port_attributes, "PortSymbolicName").strip('\x00')
    kwargs['os_device_name'] = getattr(port_attributes, "OSDeviceName").strip('\x00')
    kwargs['port_supported_speeds'] = translate_port_supported_speeds(getattr(port_attributes, "PortSuggestedSpeed"))
    kwargs['port_max_frame_size'] = getattr(port_attributes, "PortMaxFrameSize", None)
    kwargs['fabric_name'] = translate_wwn(getattr(port_attributes, "FabricName", None))
    kwargs['driver_name'] = getattr(adapter_attributes, "DriverName", '').strip('\x00')
    kwargs['driver_version'] = getattr(adapter_attributes, "DriverVersion", '').strip('\x00')
    kwargs['manufacturer'] = getattr(adapter_attributes, "Manufacturer", '').strip('\x00')
    kwargs['serial_number'] = getattr(adapter_attributes, "SerialNumber", '').strip('\x00')
    kwargs['model'] = getattr(adapter_attributes, "Model", '').strip('\x00')
    kwargs['model_description'] = getattr(adapter_attributes, "ModelDescription", '').strip('\x00')
    kwargs['hardware_version'] = getattr(adapter_attributes, "HardwareVersion", '').strip('\x00')
    kwargs['firmware_version'] = getattr(adapter_attributes, "FirmwareVersion", '').strip('\x00')
    kwargs['option_rom_version'] = getattr(adapter_attributes, "OptionROMVersion", '').strip('\x00')
    port = Port()
    for key, value in kwargs.items():
        port[key] = value
    return port



__import__("pkg_resources").declare_namespace(__name__)

#pylint: disable-all

import unittest
import mock

from .. import Generator

class UnavailableGenerator(Generator):
    @classmethod
    def is_available(cls):
        return False

class EmptyGenerator(Generator):
    @classmethod
    def is_available(cls):
        return True

    @classmethod
    def iter_ports(cls):
        for nothing in []:
            yield

def get_list_of_generators():
    return [UnavailableGenerator, EmptyGenerator]



__import__("pkg_resources").declare_namespace(__name__)

import sys
from os.path import exists, join, sep, dirname, pardir, abspath
MOCK_ROOT_FS = abspath(join(dirname(__file__), pardir, pardir, pardir, pardir, 'mock_fs', 'redhat_no_remote_ports'))

def _print_hbaapi_example():
    import json
    from infi.hbaapi import get_ports_collection
    print [port for port in get_ports_collection().get_ports()]


def hbaapi_mock(argv=sys.argv): #pylint: disable-msg=W0102,W0613
    import infi.hbaapi.generators.sysfs
    infi.hbaapi.generators.sysfs.ROOT_FS = MOCK_ROOT_FS
    _print_hbaapi_example()

def hbaapi_real(argv=sys.argv): #pylint: disable-msg=W0102,W0613
    _print_hbaapi_example()

def main(argv=sys.argv): #pylint: disable-msg=W0102,W0613
    print 'hello world'




__import__("pkg_resources").declare_namespace(__name__)

#pylint: disable-all

import logging
import unittest
import mock
from os.path import exists, join, sep, dirname, pardir, abspath
from ... import sysfs

logger = logging.getLogger(__name__)

MOCK_ROOT_FS = abspath(join(dirname(__file__), pardir, pardir, pardir, pardir, pardir, pardir, 'mock_fs', 'redhat_no_remote_ports'))
REMOTE_PORTS_MOCK_ROOT_FS = abspath(join(dirname(__file__), pardir, pardir, pardir, pardir, pardir, pardir, 'mock_fs', 'mock_fs_hbaapi-5'))

class GeneratorTestCase(unittest.TestCase):
    def test_stat_conversion__zero(self):
        self.assertEquals(0, sysfs.translate_stat_value_to_number('0x0'))

    def test_stat_conversion__ff(self):
        self.assertEquals(255, sysfs.translate_stat_value_to_number('0xff'))

    def test_stat_conversion__overflow(self):
        self.assertEquals(-1, sysfs.translate_stat_value_to_number('0xffffffffffffffff'))

    @mock.patch.object(sysfs, 'ROOT_FS' , MOCK_ROOT_FS)
    def test_mock_fs(self):
        from infi.hbaapi.tests import PortAssertions
        logger.debug("mock_fs = %s", MOCK_ROOT_FS)
        self.assertIs(sysfs.ROOT_FS, MOCK_ROOT_FS)
        self.assertTrue(sysfs.Sysfs.is_available())
        ports = [port for port in sysfs.Sysfs().iter_ports()]
        port_test_class = PortAssertions(self)
        for port in ports:
            port_test_class.assert_port(port)

    def _assert_wwn_translation(self, expected, actual):
        self.assertEquals(sysfs.translate_wwn(actual), expected)

    def test_wwn_translation(self):
        for expected, actual in [('01:02:03:04:05:06:07:08', '01:02:03:04:05:06:07:08'),
                                 ('01:02:03:04:05:06:07:08', '0x0102030405060708'),
                                 ('ab:cd:ab:cd:ab:cd:ab:cd', 'ab:cd:ab:cd:ab:cd:ab:cd'.upper())]:
            self._assert_wwn_translation(expected, actual)

class RemotePortsTestCase(unittest.TestCase):
    @mock.patch.object(sysfs, 'ROOT_FS' , REMOTE_PORTS_MOCK_ROOT_FS)
    def test_mock_fs(self):
        raise unittest.SkipTest("the mocked filesystem was removed because windows filesystem")
        logger.debug("mock_fs = %s", REMOTE_PORTS_MOCK_ROOT_FS)
        self.assertIs(sysfs.ROOT_FS, REMOTE_PORTS_MOCK_ROOT_FS)
        self.assertTrue(sysfs.Sysfs.is_available())
        self._assert_remote_ports()

    def _assert_remote_ports(self):
        from infi.dtypes.wwn import WWN
        ports = [port for port in sysfs.Sysfs().iter_ports()]
        remote_ports = ports[0].discovered_ports
        self.assertEqual(len(remote_ports), 2)
        self.assertEqual(remote_ports[0].port_wwn, WWN("5000402001f45eb5"))
        self.assertEqual(remote_ports[1].port_wwn, WWN("21000024ff2c4df3"))
        self.assertEqual(remote_ports[0].hct, (2,0,0))
        self.assertEqual(remote_ports[1].hct, (2,0,1))




__version__ = "0.1.20"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "wiggin15@yahoo.com"
__git_branch__ = '(Not currently on any branch)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '7ef2cf0e056b7507357f3a5958f11a33e3b7a34c'
__git_head_subject__ = 'Finished release v0.1.20.'
__git_head_message__ = ''
__git_dirty_diff__ = ''


from infi.hbaapi import Port
import json
example = json.loads(""" [
    {
        "discovered_ports": [], 
        "driver_name": null, 
        "driver_version": "8.03.01.05.05.06-k", 
        "fabric_name": "20:01:00:05:73:dc:fa:81", 
        "firmware_version": null, 
        "hardware_version": "5.03.02 (d5)", 
        "manufacturer": null, 
        "model": "QLE2562", 
        "model_description": "PCI-Express Dual Channel 8Gb Fibre Channel HBA", 
        "node_wwn": "20:00:00:24:ff:2c:4e:53", 
        "option_rom_version": "4.04.00 128", 
        "os_device_name": null, 
        "port_fcid": "0x4d0300", 
        "port_max_frame_size": null, 
        "port_speed": 4, 
        "port_state": "online", 
        "port_supported_speeds": [
            1, 
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "QLE2562 FW:v5.03.02 DVR:v8.03.01.05.05.06-k", 
        "port_type": "nport (fabric via point-to-point)", 
        "port_wwn": "21:00:00:24:ff:2c:4e:53", 
        "serial_number": "", 
        "statistics": {
            "dumped_frames": 0, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": 0, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": 0, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 1, 
            "lip_count": 0, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": 1, 
            "primitive_seq_protocol_err_count": -1, 
            "primitiveitive_seq_protocol_err_count": -1, 
            "rx_frames": 3475, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": 3473, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [], 
        "driver_name": null, 
        "driver_version": "8.03.01.05.05.06-k", 
        "fabric_name": "10:00:00:05:33:68:30:15", 
        "firmware_version": null, 
        "hardware_version": "5.03.02 (d5)", 
        "manufacturer": null, 
        "model": "QLE2562", 
        "model_description": "PCI-Express Dual Channel 8Gb Fibre Channel HBA", 
        "node_wwn": "20:00:00:24:ff:2c:4e:52", 
        "option_rom_version": "4.04.00 128", 
        "os_device_name": null, 
        "port_fcid": "0x011300", 
        "port_max_frame_size": null, 
        "port_speed": 8, 
        "port_state": "online", 
        "port_supported_speeds": [
            1, 
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "QLE2562 FW:v5.03.02 DVR:v8.03.01.05.05.06-k", 
        "port_type": "nport (fabric via point-to-point)", 
        "port_wwn": "21:00:00:24:ff:2c:4e:52", 
        "serial_number": "", 
        "statistics": {
            "dumped_frames": 0, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": 0, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": 0, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 4, 
            "link_failure_count": 1, 
            "lip_count": 0, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": 1, 
            "primitive_seq_protocol_err_count": -1, 
            "primitiveitive_seq_protocol_err_count": -1, 
            "rx_frames": 3474, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": 3474, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [], 
        "driver_name": null, 
        "driver_version": "8.03.01.05.05.06-k", 
        "fabric_name": "20:00:00:24:ff:2c:4d:f3", 
        "firmware_version": null, 
        "hardware_version": "5.03.02 (d5)", 
        "manufacturer": null, 
        "model": "QLE2562", 
        "model_description": "PCI-Express Dual Channel 8Gb Fibre Channel HBA", 
        "node_wwn": "20:00:00:24:ff:2c:4d:f3", 
        "option_rom_version": "4.04.00 128", 
        "os_device_name": null, 
        "port_fcid": "0x000000", 
        "port_max_frame_size": null, 
        "port_speed": 0, 
        "port_state": "online", 
        "port_supported_speeds": [
            1, 
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "QLE2562 FW:v5.03.02 DVR:v8.03.01.05.05.06-k", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4d:f3", 
        "serial_number": "", 
        "statistics": {
            "dumped_frames": 0, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": 0, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": 0, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 0, 
            "lip_count": 0, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": 0, 
            "primitive_seq_protocol_err_count": -1, 
            "primitiveitive_seq_protocol_err_count": -1, 
            "rx_frames": 0, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": 0, 
            "tx_words": -1
        }
    }, 
    {
        "discovered_ports": [], 
        "driver_name": null, 
        "driver_version": "8.03.01.05.05.06-k", 
        "fabric_name": "20:00:00:24:ff:2c:4d:f2", 
        "firmware_version": null, 
        "hardware_version": "5.03.02 (d5)", 
        "manufacturer": null, 
        "model": "QLE2562", 
        "model_description": "PCI-Express Dual Channel 8Gb Fibre Channel HBA", 
        "node_wwn": "20:00:00:24:ff:2c:4d:f2", 
        "option_rom_version": "4.04.00 128", 
        "os_device_name": null, 
        "port_fcid": "0x000000", 
        "port_max_frame_size": null, 
        "port_speed": 0, 
        "port_state": "online", 
        "port_supported_speeds": [
            1, 
            2, 
            4, 
            8
        ], 
        "port_symbolic_name": "QLE2562 FW:v5.03.02 DVR:v8.03.01.05.05.06-k", 
        "port_type": null, 
        "port_wwn": "21:00:00:24:ff:2c:4d:f2", 
        "serial_number": "", 
        "statistics": {
            "dumped_frames": 0, 
            "error_frames": -1, 
            "fcp_control_requests": -1, 
            "fcp_input_megabytes": 0, 
            "fcp_input_requests": -1, 
            "fcp_output_megabytes": 0, 
            "fcp_output_requests": -1, 
            "invalid_crc_count": 0, 
            "invalid_tx_word_count": 0, 
            "link_failure_count": 0, 
            "lip_count": 0, 
            "loss_of_signal_count": 0, 
            "loss_of_sync_count": 0, 
            "nos_count": 0, 
            "primitive_seq_protocol_err_count": -1, 
            "primitiveitive_seq_protocol_err_count": -1, 
            "rx_frames": 0, 
            "rx_words": -1, 
            "seconds_since_last_reset": -1, 
            "tx_frames": 0, 
            "tx_words": -1
        }
    }
]
""")
ports = [Port().update_not_none_values(item) for item in example]




import ctypes
import glob
import sys
from . import headers
from infi.cwrap import WrappedFunction, get_os_name, errcheck_zero, errcheck_nonzero, errcheck_nothing
from infi.cwrap import IN, IN_OUT

HBAAPI_SHARED_LIBRARY_FILENAMES = {
    'windows': 'hbaapi.dll',
    'linux': (glob.glob('/usr/lib64/libHBAAPI.so*' if sys.maxsize > 2 ** 32 \
                        else '/usr/lib/libHBAAPI.so*') + ['libHBAAPI.so'])[0]
    }

class InconsistencyError(Exception):
    pass

def errcheck_not_supported(errcheck_func=errcheck_nonzero):
    def errcheck(result, func, args):
        if result in [headers.HBA_STATUS_ERROR_NOT_SUPPORTED, headers.HBA_STATUS_ERROR_UNSUPPORTED_FC4]:
            raise NotImplementedError
        return errcheck_func()(result, func, args)
    return errcheck

def errcheck_inconsistency(errcheck_func=errcheck_nonzero):
    def errcheck(result, func, args):
        if result in [headers.HBA_STATUS_ERROR_STALE_DATA, headers.HBA_STATUS_ERROR_INVALID_HANDLE,
                      headers.HBA_STATUS_ERROR_ILLEGAL_INDEX, headers.HBA_STATUS_ERROR_TRY_AGAIN,
                      headers.HBA_STATUS_ERROR_UNAVAILABLE]:
            raise InconsistencyError("%s raised return an inconsistency error %s, you should retry",
                                     func.__name__, result)
        return errcheck_func()(result, func, args)
    return errcheck

class HbaApiFunction(WrappedFunction):
    @classmethod
    def _get_function_type(cls):
        return 'CFUNCTYPE'

    @classmethod
    def _get_library(cls):
        library_name = HBAAPI_SHARED_LIBRARY_FILENAMES.get(get_os_name(), None)
        if library_name is None:
            raise OSError
        try:
            return ctypes.cdll.LoadLibrary(library_name)
        except:
            raise OSError

# Library Control Functions

class HBA_GetVersion(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_UINT32

    @classmethod
    def get_errcheck(cls):
        return errcheck_zero()

    @classmethod
    def get_parameters(cls):
        return ()

class HBA_LoadLibrary(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_parameters(cls):
        return ()

class HBA_FreeLibrary(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()
    @classmethod
    def get_parameters(cls):
        return ()

class HBA_GetNumberOfAdapters(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_UINT32

    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()

    @classmethod
    def get_parameters(cls):
        return ()

class HBA_RefreshAdapterConfiguration(HbaApiFunction): #pylint: disable-msg=C0103
    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()

    @classmethod
    def get_parameters(cls):
        return ()

# Adapter and Port Information Functions

class HBA_GetAdapterName(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()

    @classmethod
    def get_parameters(cls):
        return (headers.UINT32, IN, 'adapterIndex'), \
            (ctypes.POINTER(ctypes.c_char), IN, 'adapterName')

class HBA_OpenAdapter(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_HANDLE

    @classmethod
    def get_errcheck(cls):
        return errcheck_zero()

    @classmethod
    def get_parameters(cls):
        return (ctypes.POINTER(ctypes.c_char), IN, 'adapterName'),

class HBA_CloseAdapter(HbaApiFunction): #pylint: disable-msg=C0103
    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()

    @classmethod
    def get_parameters(cls):
        return (headers.HBA_HANDLE, IN_OUT, 'handle'),

class HBA_GetAdapterAttributes(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_inconsistency(errcheck_not_supported)

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, 'handle'),
                (ctypes.c_void_p, IN_OUT, 'adapterAttributes'))

class HBA_GetAdapterPortAttributes(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_inconsistency(errcheck_not_supported)

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, 'handle'),
                (headers.HBA_UINT32, IN, 'portIndex'),
                (ctypes.c_void_p, IN_OUT, 'portAttributes'))

class HBA_GetDiscoveredPortAttributes(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_inconsistency(errcheck_not_supported)

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, 'handle'),
                (headers.HBA_UINT32, IN, 'portIndex'),
                (headers.HBA_UINT32, IN, 'discoveredPortIndex'),
                (ctypes.c_void_p, IN_OUT, 'portAttributes'))

class HBA_GetPortStatistics(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_inconsistency(errcheck_not_supported)

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, 'handle'),
                (headers.HBA_UINT32, IN, 'portIndex'),
                (ctypes.c_void_p, IN_OUT, 'portStatistics'))

class HBA_GetFC4Statistics(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_inconsistency(errcheck_not_supported)

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, 'handle'),
                (ctypes.c_uint64, IN, 'portWWN'),
                (ctypes.c_uint8, IN, 'FC4type'),
                (ctypes.c_void_p, IN_OUT, 'fcpStatistics'))

class HBA_GetFcpTargetMappingV2(HbaApiFunction): #pylint: disable-msg=C0103
    return_value = headers.HBA_STATUS

    @classmethod
    def get_errcheck(cls):
        return errcheck_nonzero()

    @classmethod
    def get_parameters(cls):
        return ((headers.HBA_HANDLE, IN, "handle"),
                (ctypes.c_uint64, IN, "hbaPortWWN"),
                (ctypes.c_void_p, IN_OUT, "pMapping"))




__import__("pkg_resources").declare_namespace(__name__)

class Generator(object):
    def __init__(self):
        object.__init__(self)

    def iter_ports(self):
        raise NotImplementedError

    @classmethod
    def is_available(cls):
        raise NotImplementedError

class CompositeGenerator(Generator):
    def __init__(self):
        self._children = []

    def iter_ports(self):
        for child in self._children:
            for port in child().iter_ports():
                yield port

    @classmethod
    def is_available(cls):
        return True

    def add_generator(self, generator):
        if generator not in self._children:
            self._children.append(generator)

    def remove_generator(self, generator):
        self._children.remove(generator)

    def iter_generators(self):
        for generator in self._children:
            yield generator()

def get_list_of_generators():
    from .hbaapi import HbaApi
    from .sysfs import Sysfs
    return [HbaApi, Sysfs]




import ctypes
from infi.instruct import Struct, VarSizeArray, ReadPointer, Field, Padding
from infi.instruct import FixedSizeArray as Array
from infi.instruct import FixedSizeString as String
from infi.instruct import UNInt32, UNInt8, SNInt64, UNInt64

##############
# Data Types #
##############
UINT32 = ctypes.c_uint32

HBA_UINT8 = ctypes.c_uint8
HBA_UINT32 = ctypes.c_uint32
HBA_HANDLE = HBA_UINT32
HBA_BOOLEAN = HBA_UINT8
HBA_UINT8 = HBA_BOOLEAN
HBA_STATUS = HBA_UINT32
HBA_PORTTYPE = HBA_UINT32
HBA_PORTSTATE = HBA_UINT32
HBA_PORTSPEED = HBA_UINT32
HBA_COS = HBA_UINT32

##############
# Structures #
##############

BitsArray = Array('bits', 32, UNInt8)

def HBA_WWN(name):
    return String(name, 8)

PortSupportedFc4Types = Array("PortSupportedFc4Types", 32, UNInt8)
PortActiveFc4Types = Array("PortActiveFc4Types", 32, UNInt8)

def is_64bit():
    from sys import maxsize
    return maxsize > 2 ** 32

class HBA_AdapterAttributes(Struct): #pylint: disable-msg=C0103
    _fields_ = [
               String("Manufacturer", 64),
               String("SerialNumber", 64),
               String("Model", 256),
               String("ModelDescription", 256),
               HBA_WWN("NodeWWN"),
               String("NodeSymbolicName", 256),
               String("HardwareVersion", 256),
               String("DriverVersion", 256),
               String("OptionROMVersion", 256),
               String("FirmwareVersion", 256),
               UNInt32("VendorSpecificID"),
               UNInt32("NumberOfPorts"),
               String("DriverName", 256)
               ]

class HBA_PortAttributes(Struct): #pylint: disable-msg=C0103
    _fields_ = [
                HBA_WWN("NodeWWN"),
                HBA_WWN("PortWWN"),
                UNInt32("PortFcId"),
                UNInt32("PortType"),
                UNInt32("PortState"),
                UNInt32("PortSupportedClassofService"),
                PortSupportedFc4Types,
                PortActiveFc4Types,
                String("PortSymbolicName", 256),
                String("OSDeviceName", 256),
                UNInt32("PortSuggestedSpeed"),
                UNInt32("PortSpeed"),
                UNInt32("PortMaxFrameSize"),
                HBA_WWN("FabricName"),
                UNInt32("NumberOfDiscoveredPorts")
                ]

class HBA_PortStatistics(Struct): #pylint: disable-msg=C0103
    _fields_ = [
                SNInt64("SecondsSinceLastReset"),
                SNInt64("TxFrames"),
                SNInt64("TxWords"),
                SNInt64("RxFrames"),
                SNInt64("RxWords"),
                SNInt64("LIPCount"),
                SNInt64("NOSCount"),
                SNInt64("ErrorFrames"),
                SNInt64("DumpedFrames"),
                SNInt64("LinkFailureCount"),
                SNInt64("LossOfSyncCount"),
                SNInt64("LossOfSignalCount"),
                SNInt64("PrimitiveSeqProtocolErrCount"),
                SNInt64("InvalidTxWordCount"),
                SNInt64("InvalidCRCCount")
                ]

class HBA_FC4Statistics(Struct): #pylint: disable-msg=C0103
    _fields_ = [
               SNInt64("InputRequests"),
               SNInt64("OutputRequests"),
               SNInt64("ControlRequests"),
               SNInt64("InputMegabytes"),
               SNInt64("OutputMegabytes")
               ]

class HBA_SCSIID(Struct): #pylint: disablemsg=C0103
    _fields_ = [
                String("OSDeviceName", 256),
                UNInt32("ScsiBusNumber"),
                UNInt32("ScsiTargetNumber"),
                UNInt32("ScsiOSLun"),
                Padding(4), ]

class HBA_FcpId(Struct): #pylint: disablemsg=C0103
    _fields_ = [
                UNInt32("Fcid"),
                HBA_WWN("NodeWWN"),
                HBA_WWN("PortWWN"),
                UNInt64("FcpLun"),
                ]

class HBA_LUID(Struct): #pylint: disablemsg=C0103
    _fields_ = [
                String("buffer", 256),
                ]

class HBA_FcpScsiEntryV2(Struct): #pylint: disablemsg=C0103
    _fields_ = [Field("ScsiId", HBA_SCSIID),
                Field("FcId", HBA_FcpId),
                Field("LUID", HBA_LUID),
                Padding(4),
                ]

class HBA_FCPTargetMappingV2(Struct): #pylint: disablemsg=C0103
    _fields_ = [
                UNInt32("NumberOfEntries"),
                Padding(4),
                VarSizeArray("entry", ReadPointer("NumberOfEntries"), HBA_FcpScsiEntryV2)
                ]

#############
# Constants #
#############

HBA_STATUS_OK = 0
HBA_STATUS_ERROR = 1
HBA_STATUS_ERROR_NOT_SUPPORTED = 2
HBA_STATUS_ERROR_INVALID_HANDLE = 3
HBA_STATUS_ERROR_ARG = 4
HBA_STATUS_ERROR_ILLEGAL_WWN = 5
HBA_STATUS_ERROR_ILLEGAL_INDEX = 6
HBA_STATUS_ERROR_MORE_DATA = 7
HBA_STATUS_ERROR_STALE_DATA = 8
HBA_STATUS_SCSI_CHECK_CONDITION = 9
HBA_STATUS_ERROR_BUSY = 10
HBA_STATUS_ERROR_TRY_AGAIN = 11
HBA_STATUS_ERROR_UNAVAILABLE = 12
HBA_STATUS_ERROR_ELS_REJECT = 13
HBA_STATUS_ERROR_INVALID_LUN = 14
HBA_STATUS_ERROR_INCOMPATIBLE = 15
HBA_STATUS_ERROR_AMBIGUOUS_WWN = 16
HBA_STATUS_ERROR_LOCAL_BUS = 17
HBA_STATUS_ERROR_LOCAL_TARGET = 18
HBA_STATUS_ERROR_LOCAL_LUN = 19
HBA_STATUS_ERROR_LOCAL_SCSIID_BOUND = 20
HBA_STATUS_ERROR_TARGET_FCID = 21
HBA_STATUS_ERROR_TARGET_NODE_WWN = 22
HBA_STATUS_ERROR_TARGET_PORT_WWN = 23
HBA_STATUS_ERROR_TARGET_LUN = 24
HBA_STATUS_ERROR_TARGET_LUID = 25
HBA_STATUS_ERROR_NO_SUCH_BINDING = 26
HBA_STATUS_ERROR_NOT_A_TARGET = 27
HBA_STATUS_ERROR_UNSUPPORTED_FC4 = 28
HBA_STATUS_ERROR_INCAPABLE = 29

HBA_PORTTYPE_UNKNOWN = 1
HBA_PORTTYPE_OTHER = 2
HBA_PORTTYPE_NOTPRESENT = 3
HBA_PORTTYPE_NPORT = 5
HBA_PORTTYPE_NLPORT = 6
HBA_PORTTYPE_FLPORT = 7
HBA_PORTTYPE_FPORT = 8
HBA_PORTTYPE_EPORT = 9
HBA_PORTTYPE_GPORT = 10
HBA_PORTTYPE_LPORT = 20
HBA_PORTTYPE_PTP = 21

HBA_PORTTYPE = {'0':None,
                '1':None,
                '2':None,
                '3':None,
                '5':'N',
                '6':'NL',
                '7':'FL',
                '8':'F',
                '9':'E',
                '10':'G',
                '20':'L',
                '21':'PTP'}

HBA_PORTSTATE_UNKNOWN = 1
HBA_PORTSTATE_ONLINE = 2
HBA_PORTSTATE_OFFLINE = 3
HBA_PORTSTATE_BYPASSED = 4
HBA_PORTSTATE_DIAGNOSTICS = 5
HBA_PORTSTATE_LINKDOWN = 6
HBA_PORTSTATE_ERROR = 7
HBA_PORTSTATE_LOOPBACK = 8

HBA_PORTSPEED_UNKNOWN = 0
HBA_PORTSPEED_1GBIT = 1
HBA_PORTSPEED_2GBIT = 2
HBA_PORTSPEED_10GBIT = 4
HBA_PORTSPEED_4GBIT = 8
HBA_PORTSPEED_8GBIT = 16
HBA_PORTSPEED_NOT_NEGOTIATED = 1 << 15

MAX_ADAPTERNAME_LENGTH = 256
